package gui;

import java.io.File;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;
import javax.swing.AbstractAction;
import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.Position;
import javax.swing.text.StyledDocument;

/**
 *
 */
public class Log {

  private static DefaultStyledDocument document;
  private static List<JTextPane> textPanes;
  private static Indent indent;
  private static boolean traceCalls;
  private static StringBuilder theStringBuilder;
  private static int callCount;
  private static JFileChooser fileChooser;
  private static Action saveAction;
  private static Action clearAction;
  private static Action setTraceCallsAction;

  /**
   * getDocument() returns the document that is shared by all Log view
   * components.
   * This document is created if it does not yet exist.
   * @return the document shared by all Log view components
   */
  private static StyledDocument getDocument() {
    if (document == null) {
      document = new DefaultStyledDocument();
      textPanes = new ArrayList<JTextPane>();
    }
    return document;
  }

  /**
   * getIndent() returns the indent object used in this log.
   * This indent object is created if it does not yet exist.
   * @return the indent object
   */
  private static Indent getIndent() {
    if (indent == null) {
      indent = new Indent(2);
    }
    return indent;
  }

  /**
   * getStringBuilder() returns the string builder used in this log.
   * This string builder is created if it does not yet exist.
   * Otherwise it is cleared.
   * @return the string builder
   */
  private static StringBuilder getStringBuilder() {
    if (theStringBuilder == null) {
      theStringBuilder = new StringBuilder();
    }
    theStringBuilder.setLength(0);
    return theStringBuilder;
  }

  /**
   * getStringBuilder(initialText) returns the string builder used in this log.
   * This string builder is created if it does not yet exist.
   * @return the string builder
   */
  private static StringBuilder getStringBuilder(String initialText) {
    StringBuilder stringBuilder = getStringBuilder();
    stringBuilder.append(initialText);
    return stringBuilder;
  }

  private static void appendTextLine(String logText) {
    Position end = getDocument().getEndPosition();
    try {
      document.insertString(end.getOffset() - 1, logText + '\n', null);
      for (JTextPane textPane : textPanes) {
        textPane.setCaretPosition(textPane.getText().length());
      }
    } catch (BadLocationException exception) {
      // This should never happen.
      System.out.println(exception.getMessage());
    }
  }

  /**
   * Log.createViewComponent() returns a new Log view component.
   * The view component is a scroll pane containing a text pane.
   * The text pane displays logged text.
   * @return a new Log view component
   */
  public static JComponent createViewComponent() {
    if (textPanes == null) {
      textPanes = new ArrayList<JTextPane>();
    }
    JTextPane textPane = new JTextPane();
    textPanes.add(textPane);
    textPane.setDocument(getDocument());
    textPane.setEditable(false);
    textPane.setPreferredSize(new Dimension(100, 200));
    JScrollPane scrollPane = new JScrollPane(textPane);
    scrollPane.setHorizontalScrollBarPolicy(
            JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    scrollPane.setVerticalScrollBarPolicy(
            JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
    scrollPane.setBorder(BorderFactory.createTitledBorder("Log"));
    return scrollPane;
  }

  /**
   * Log.reset() removes all log entries from Log and resets Log.
   */
  public static void clear() {
    reset();
    try {
      document.remove(0, document.getLength());
    } catch (BadLocationException exception) {
      // This should never happen.
      System.out.println(exception.getMessage());
    }
  }

  /**
   * Log.log(text) adds a log entry for text to Log.
   * The current log indent is added at the beginning of the entry.
   * A newline character is automatically added to the end of the entry.
   * @param text the text to be added
   */
  public static void log(String text) {
    appendTextLine(getIndent().toString() + text);
  }

  /**
   * Log.log(object) adds a log entry describing object to Log.
   * The description is obtained by invoking object.toString().
   * The current log indent is added at the beginning of the entry.
   * A newline character is automatically added to the end of the entry.
   * @param object the text to be added
   */
  public static void log(Object object) {
    appendTextLine(getIndent().toString() + object);
  }

  /**
   * Log.log() adds a blank line log entry to Log.
   */
  public static void log() {
    appendTextLine("");
  }

  /**
   * Log.reset() initializes the call count to 0 and resets the
   * indentation level.
   */
  public static void reset() {
    callCount = 0;
    getIndent().reset();
  }

  /**
   * Log.getCallCount() returns the call count.
   * @return the call count
   */
  public static int getCallCount() {
    return callCount;
  }

  /**
   * Log.setTraceCalls(newTraceCalls) sets the traceCalls attribute of Log to
   * newTraceCalls.
   * If doTraceCalls is true, future enter() and leave() message entries will
   * be added to Log.
   * Otherwise they will be ignored except for call counting.
   * @param newTraceCalls the new value for the traceCalls attribute
   */
  public static void setTraceCalls(boolean newTraceCalls) {
    traceCalls = newTraceCalls;
  }

  /**
   * Log.enter(methodName, parameters...) adds a log entry for entry into a
   * method named methodName.
   * The parameters for the entered method can be included as additional
   * parameters for Log.enter().
   * These additional parameters are printed using Java default conversions
   * to string values.
   * @param methodName the name of the method
   * @param parameters zero or more method parameters
   */
  public static void enter(String methodName, Object... parameters) {
    if (traceCalls) {
      StringBuilder stringBuilder = getStringBuilder(indent.toString());
      stringBuilder.append("invoking ");
      stringBuilder.append(methodName);
      stringBuilder.append('(');
      stringBuilder.append(parameters[0]);
      for (int i = 1; i < parameters.length; i++) {
        stringBuilder.append(", ");
        stringBuilder.append(parameters[i]);
      }
      stringBuilder.append(')');
      appendTextLine(stringBuilder.toString());
      indent.enter();
    }
    callCount++;
  }

  /**
   * Log.leave(returnValue) decrements the level attribute of Log and then adds
   * a method return log entry.
   * The log entry indicates that the method in which leave() was invoked
   * returned returnValue.
   * @param returnValue
   */
  public static void leave(Object returnValue) {
    if (traceCalls) {
      indent.leave();
      appendTextLine(indent.toString() + "return " + returnValue);
    }
  }

  /**
   * Log.leave() decrements the level attribute of Log and then adds a method
   * return log entry.
   * The log entry indicates that the method in which leave() was invoked
   * returned without a return value.
   */
  public static void leave() {
    if (traceCalls) {
      indent.leave();
      appendTextLine(indent.toString() + "return");
    }
  }

  /**
   * Log.getContents() returns the contents of the log.
   * @return
   */
  public static String getContents() {
    try {
      return getDocument().getText(0, document.getLength());
    } catch (Exception ex) {
      return ex.toString();
    }
  }

  /**
   * Log.save() brings up a file dialog for saving the contents of the log.
   */
  public static void save() {
    if (fileChooser == null) {
      fileChooser = new JFileChooser();
    }
    int choice = fileChooser.showSaveDialog(fileChooser);
    if (choice != JFileChooser.APPROVE_OPTION) {
      return;
    }
    File file = fileChooser.getSelectedFile();
    try {
      FileWriter writer = new FileWriter(file);
      writer.append(getContents());
      writer.close();
    } catch (Exception exception) {
      JOptionPane.showMessageDialog(null, exception.getMessage(),
              "Save Error", JOptionPane.ERROR_MESSAGE);
    }
  }

  /**
   * Log.getSaveAction() returns the save action.
   * This action brings up a file dialog for saving the contents of the log.
   * @return the save action
   */
  public static Action getSaveAction() {
    if (saveAction == null) {
      saveAction = new AbstractAction("Save Log") {

        @Override
        public void actionPerformed(ActionEvent ev) {
          save();
        }

      };
    }
    return saveAction;
  }

  /**
   * Log.getSaveAction() returns the clear action.
   * This action clears the contents of the log.
   * @return the clear action
   */
  public static Action getClearAction() {
    if (clearAction == null) {
      clearAction = new AbstractAction("Clear Log") {

        @Override
        public void actionPerformed(ActionEvent ev) {
          clear();
        }

      };
    }
    return clearAction;
  }

  /**
   * Log.getTraceCallsAction() returns the trace calls action.
   * This action sets the trace calls attribute of Log to the state of the
   * button it is installed on.
   * @return the trace calls action
   */
  public static Action getSetTraceCallsAction() {
    if (setTraceCallsAction == null) {
      setTraceCallsAction = new AbstractAction("Trace Calls") {

        @Override
        public void actionPerformed(ActionEvent event) {
          AbstractButton button = (AbstractButton) event.getSource();
          setTraceCalls(button.isSelected());
        }

      };
    }
    return setTraceCallsAction;
  }

}
