package totalprice;

import gui.Log;
import java.util.HashMap;
import java.util.Map;

/**
 * A MemoizedSimpleTotalPriceOptimizer computes the largest total price that can
 * be obtained by cutting a bar into sub-bars of various lengths.
 * The price available for a sub-bar is determined by invoking static methods of
 * the PriceList class.
 *
 * This implementation tries all possible lengths for the first cut and recurs
 * for the remainder.
 * It uses memoization to avoid repeated solutions of the same subproblem.
 */
public class MemoizedSimpleTotalPriceOptimizer
        implements TotalPriceOptimizer {

  /**
   * A Parameters object represents the parameters for a call to
   * bestTotalPrice().
   * It is used in memoization as a key for the solutions table.
   *
   * This implementation is designed for use with a hash table.
   * For use with ordered tables such as java.util.TreeMap, the Comparable
   * interface should be implemented by the Parameters class.
   */
  public class Parameters {

    private int length;

    /**
     * new Parameters(length) returns a new Parameters object representing
     * bestTotalPrice() parameter value length.
     * @param length the bestTotalPrice() parameter
     */
    private Parameters(int length) {
      this.length = length;
    }

    /**
     * param.equals(obj) returns true if obj is a Parameters object with the
     * same parameter values as param.
     * @param obj another object
     * @return true if the other object has the same parameter values
     */
    @Override
    public boolean equals(Object obj) {
      if (!(obj instanceof Parameters)) {
        return false;
      }
      Parameters parameters = (Parameters) obj;
      return parameters.length == length;
    }

    /**
     * param.hashCode() returns the hash code for param.
     * Two Parameters objects that are equal must return the same hash code.
     * @return the hash code
     */
    @Override
    public int hashCode() {
      return length;
    }

  }

  private PriceList priceList;
  // Memoization requires a table for looking up solutions of subproblems that
  // have already been solved.
  private Map<Parameters, Double> solutions;

  private double doBestTotalPrice(int length) {
    Log.enter("doBestTotalPrice", length);
    double bestYetTotalPrice = 0.0;
    // The next 5 lines are added for memoization.
    // They avoid resolving a subproblem that has already been solved.
    Parameters parameters = new Parameters(length);
    Double solution = solutions.get(parameters);
    if (solution != null) {
      Log.leave(solution);
      return solution;
    }
    for (int newLength = 1; newLength <= length; newLength++) {
      // Run time bookeeping:
      // Don't count the loop maintenance time for this call as part of its
      // run time.
      // Instead, count loop maintenance overhead for each iteration as part of
      // the run time for the recursive call in that iteration.
      // Then total run time is bounded by a constant times the call count.
      double newTotalPrice = priceList.getPrice(newLength);
      if (newLength < length) {
        newTotalPrice += doBestTotalPrice(length - newLength);
      }
      if (newTotalPrice > bestYetTotalPrice) {
        bestYetTotalPrice = newTotalPrice;
      }
    }
    // For memoization, all solutions should be recorded in the solutions table.
    solutions.put(parameters, bestYetTotalPrice);
    Log.leave(bestYetTotalPrice);
    return bestYetTotalPrice;
  }

  /**
   * new MemoizedSimpleTotalPriceOptimizer(priceList) returns a new memoized
   * simple total price optimizer.
   * It uses priceList to determine the price for sub-bars.
   * @param priceList the price list for sub-bars
   */
  public MemoizedSimpleTotalPriceOptimizer(PriceList priceList) {
    this.priceList = priceList;
    solutions = new HashMap<Parameters, Double>();
  }

  /**
   * tpo.bestTotalPrice(length) returns the best total price that can be charged
   * for sub-bars cut from a bar of length length.
   * @param length the bar length
   * @return the best total price
   */
  @Override
  public double bestTotalPrice(int length) {
    Log.reset();
    solutions.clear();
    return doBestTotalPrice(length);
  }

  /**
   * tpo.toString() returns a string description for tpo.
   * @return a string description
   */
  @Override
  public String toString() {
    return "Memoized Simple Total Price Optimizer";
  }

  /**
   * @param args ignored
   */
  public static void main(String[] args) {
    Log.setTraceCalls(false);
    PriceList priceList = new PriceList(1.0, 10);
    Log.log(priceList.toString());
    Log.log();
    MemoizedSimpleTotalPriceOptimizer optimizer =
            new MemoizedSimpleTotalPriceOptimizer(priceList);
    Log.log(optimizer);
    Log.log("Best Total Prices:");
    // To keep total run time under 30 seconds, use maxLength <= 1000.
    // Stick to maxLength <= 20 if tracing calls.
    int maxLength = 20;
    for (int length = 1; length <= maxLength; length++) {
      Log.log(length + ": " + optimizer.bestTotalPrice(length)
              + ", call count = " + Log.getCallCount());
    }
    System.out.println(Log.getContents());
  }

}
