package totalprice;

import gui.Log;
import java.util.HashMap;
import java.util.Map;

/**
 * A MemoizedSimpleTotalPriceSolver determines how to cut a bar into sub-bars in
 * order to get the largest total price.
 * The price list for sub-bars is specified by a constructor parameter.
 *
 * This implementation tries all possible lengths for the first cut and recurs
 * for the remainder.
 * It uses memoization to avoid repeated solutions of the same subproblem.
 */
public class MemoizedSimpleTotalPriceSolver
        implements TotalPriceSolver {

  /**
   * A Parameters object represents the parameters for a call to
   * bestTotalPrice().
   * It is used in memoization as a key for the solutions table.
   *
   * This implementation is designed for use with a hash table.
   * For use with ordered tables such as java.util.TreeMap, the Comparable
   * interface should be implemented by the Parameters class.
   */
  public class Parameters {

    private int length;

    /**
     * new Parameters(length) returns a new Parameters object representing
     * bestTotalPrice() parameter value length.
     * @param length the bestTotalPrice() parameter
     */
    private Parameters(int length) {
      this.length = length;
    }

    /**
     * param.equals(obj) returns true if obj is a Parameters object with the
     * same parameter values as param.
     * @param obj another object
     * @return true if the other object has the same parameter values
     */
    @Override
    public boolean equals(Object obj) {
      if (!(obj instanceof Parameters)) {
        return false;
      }
      Parameters parameters = (Parameters) obj;
      return parameters.length == length;
    }

    /**
     * param.hashCode() returns the hash code for param.
     * Two Parameters objects that are equal must return the same hash code.
     * @return the hash code
     */
    @Override
    public int hashCode() {
      return length;
    }

  }

  private PriceList priceList;
  private Map<Parameters, Solution> solutions;

  /**
   * tpo.bestSolution(length) returns the best solution for cutting a bar of
   * length length.
   * The best solution maximizes the total price for the sub-bars.
   * @param length the bar length
   * @return the best solution
   */
  private Solution bestSolution(int length) {
    Log.enter("bestSolution", length);
    Parameters parameters = new Parameters(length);
    Solution solution = solutions.get(parameters);
    if (solution != null) {
      Log.leave(solution);
      return solution;
    }
    Solution bestYetSolution = new Solution(priceList, length);
    for (int newLength = length - 1; newLength > 0; newLength--) {
      Solution newSubproblemSolution = bestSolution(length - newLength);
      bestYetSolution.reviseSolution(newLength, newSubproblemSolution);
    }
    solutions.put(parameters, bestYetSolution);
    Log.leave(bestYetSolution);
    return bestYetSolution;
  }

  /**
   * new MemoizedSimpleTotalPriceSolver(priceList) returns a total price solver.
   * It uses priceList to determine the price for sub-bars.
   * @param priceList the price list for sub-bars
   */
  public MemoizedSimpleTotalPriceSolver(PriceList priceList) {
    this.priceList = priceList;
    solutions = new HashMap<Parameters, Solution>();
  }

  /**
   * tpo.getBestSolution(length) returns the best solution for cutting sub-bars
   * from a bar of length length.
   * The best solution is the one that yields the best total price.
   * @param length the length of a bar
   * @return the best solution for cutting the bar
   */
  @Override
  public Solution getBestSolution(int length) {
    Log.reset();
    solutions.clear();
    return bestSolution(length);
  }

  /**
   * tpo.toString() returns a string description for tpo.
   * @return a string description
   */
  @Override
  public String toString() {
    return "Memoized Simple Total Price Solver";
  }

  /**
   * @param args ignored
   */
  public static void main(String[] args) {
    Log.setTraceCalls(false);
    PriceList priceList = new PriceList(1.0, 10);
    Log.log(priceList);
    Log.log();
    MemoizedSimpleTotalPriceSolver solver =
            new MemoizedSimpleTotalPriceSolver(priceList);
    Log.log(solver);
    Log.log("Best Solutions:");
    // To keep total run time under 30 seconds, use maxLength <= 100.
    // Stick to maxLength <= 20 if tracing calls.
    int maxLength = 20;
    for (int length = 1; length <= maxLength; length++) {
      Log.log(solver.getBestSolution(length));
    }
    System.out.println(Log.getContents());
  }

}
