package totalprice;

import gui.Log;

/**
 * A OrderedTotalPriceOptimizer computes the largest total price that can be
 * obtained by cutting a bar into sub-bars of various lengths.
 * The price available for a sub-bar is determined by invoking static methods of
 * the PriceList class.
 *
 * This implementation tries all possible lengths for the first cut and recurs
 * for the remainder, with the first cut length used as a bound for recursive
 * cuts.
 */
public class OrderedTotalPriceOptimizer
        implements TotalPriceOptimizer {

  private PriceList priceList;

  /**
   * bestTotalPrice(length, maxLength) returns the best total price that can be
   * obtained from a bar of length length using sub-bar lengths that do not
   * exceed maxLength.
   * @param length the bar length
   * @param maxLength the maximum sub-bar length
   * @return the best total price
   */
  private double bestTotalPrice(int length, int maxLength) {
    Log.enter("bestTotalPrice", length, maxLength);
    double bestYetTotalPrice = 0.0;
    if (maxLength == 1) {
      bestYetTotalPrice = length * priceList.getPrice(1);
      Log.leave(bestYetTotalPrice);
      return bestYetTotalPrice;
    }
    for (int newLength = 1; newLength <= maxLength; newLength++) {
      // Run time bookeeping:
      // Don't count the loop maintenance time for this call as part of its
      // run time.
      // Instead, count loop maintenance overhead for each iteration as part of
      // the run time for the recursive call in that iteration.
      // Then total run time is bounded by a constant times the call count.
      int remainderLength = length - newLength;
      int newMaxLength =
              remainderLength < newLength ? remainderLength : newLength;
      double newTotalPrice = priceList.getPrice(newLength);
      if (remainderLength > 0) {
        newTotalPrice += bestTotalPrice(remainderLength, newMaxLength);
      }
      bestYetTotalPrice = Math.max(bestYetTotalPrice, newTotalPrice);
    }
    Log.leave(bestYetTotalPrice);
    return bestYetTotalPrice;
  }

  /**
   * new OrderedTotalPriceOptimizer(priceList) returns an ordered total price
   * optimizer.
   * It uses priceList to determine the price for sub-bars.
   * @param priceList the price list for sub-bars
   */
  public OrderedTotalPriceOptimizer(PriceList priceList) {
    this.priceList = priceList;
  }

  /**
   * tpo.bestTotalPrice(length) returns the best total price that can be charged
   * for sub-bars cut from a bar of length length.
   * @param length the bar length
   * @return the best total price
   */
  @Override
  public double bestTotalPrice(int length) {
    Log.reset();
    return bestTotalPrice(length, length);
  }

  /**
   * tpo.toString() returns a string description for tpo.
   * @return a string description
   */
  @Override
  public String toString() {
    return "Ordered Total Price Optimizer";
  }

  /**
   * @param args ignored
   */
  public static void main(String[] args) {
    Log.setTraceCalls(false);
    PriceList priceList = new PriceList(1.0, 10);
    Log.log(priceList.toString());
    Log.log();
    OrderedTotalPriceOptimizer optimizer =
            new OrderedTotalPriceOptimizer(priceList);
    Log.log(optimizer);
    Log.log("Best Total Prices:");
    // To keep total run time under 30 seconds, use maxLength <= 80.
    // Stick to maxLength <= 20 if tracing calls.
    int maxLength = 20;
    for (int length = 1; length <= maxLength; length++) {
      Log.log(length + ": " + optimizer.bestTotalPrice(length)
              + ", call count = " + Log.getCallCount());
    }
    System.out.println(Log.getContents());
  }

}
