package totalprice;

import java.util.Arrays;
import java.util.Random;

/**
 * The PriceList class provides a getPrice() method for determining the
 * price that can be charged for bars of various lengths.
 */
public class PriceList {

  private double pricePerInch;
  private int maxLength;
  private double[] prices;
  private String description;

  private void createDescription() {
    StringBuilder stringBuilder = new StringBuilder("PriceList:");
    for (int i = 1; i < prices.length; i++) {
      stringBuilder.append("\n");
      stringBuilder.append(i);
      stringBuilder.append(": ");
      stringBuilder.append(getPrice(i));
    }
    description = stringBuilder.toString();
  }

  /**
   * createRandomPrices() creates and initializes prices randomly.
   */
  private void createRandomPrices() {
    prices = new double[maxLength + 1];
    prices[0] = 0.0;
    Random rng = new Random(111111);
    for (int length = 0; length <= maxLength; length++) {
      prices[length] = pricePerInch * length * (0.5 + rng.nextDouble());
    }
    createDescription();
  }

  /**
   * createPricesFromArray(priceArray) copies prices from priceArray except
   * changing the entry with index 0 to 0.0.
   * @param priceArray the price array to be copied
   */
  private void createPricesFromArray(double[] priceArray) {
    prices = Arrays.copyOf(prices, priceArray.length);
    prices[0] = 0.0;
    maxLength = priceArray.length - 1;
    createDescription();
  }

  /**
   * new PriceList(pricePerInch, maxlength) returns a random price list.
   * The price for a bar of length l &lt;= maxlength is uniformly distributed
   * between 0.5*l*pricePerInch and 1.5*l*pricePerInch.
   * The price for a bar of length l &gt; maxLength is l*pricePerInch.
   * @param pricePerInch the average price per inch
   * @param maxLength the cutoff beyond which a flat rate per inch is used
   */
  public PriceList(double pricePerInch, int maxLength) {
    this.pricePerInch = pricePerInch;
    this.maxLength = maxLength;
    createRandomPrices();
  }

  /**
   * new PriceList(priceArray, pricePerInch) returns a price list determined
   * from priceList.
   * The price for bar lengths up to but not including priceArray.length is
   * determined by priceArray.
   * The price for a bar of length l &gt;= priceArray.length is l*pricePerInch.
   * @param pricePerInch
   * @param maxLength
   */
  public void PriceList(double[] priceArray, double pricePerInch) {
    this.pricePerInch = pricePerInch;
    createPricesFromArray(priceArray);
  }

  /**
   * pl.getPrice(length) returns the price that can be charged for a bar of
   * length length.
   * @param length the bar length in inches
   * @return the price that can be charged
   */
  public double getPrice(int length) {
    if (length <= maxLength) {
      return prices[length];
    }
    return length * pricePerInch;
  }

  /**
   * pl.toString() returns a description of pl.
   * @return a description
   */
  @Override
  public String toString() {
    return description;
  }

}
