package totalprice;

import gui.Log;

/**
 * A SimpleTotalPriceOptimizer computes the largest total price that can be
 * obtained by cutting a bar into sub-bars of various lengths.
 * The price available for a sub-bar is determined by invoking static methods of
 * the PriceList class.
 *
 * This implementation tries all possible lengths for the first cut and recurs
 * for the remainder.
 */
public class SimpleTotalPriceOptimizer
        implements TotalPriceOptimizer {

  private PriceList priceList;

  private double doBestTotalPrice(int length) {
    Log.enter("doBestTotalPrice", length);
    double bestYetTotalPrice = 0.0;
    for (int newLength = length; newLength > 0; newLength--) {
      // Run time bookeeping:
      // Don't count the loop maintenance time for this call as part of its
      // run time.
      // Instead, count loop maintenance overhead for each iteration as part of
      // the run time for the recursive call in that iteration.
      // Then total run time is bounded by a constant times the call count.
      double newTotalPrice = priceList.getPrice(newLength);
      if (newLength < length) {
        newTotalPrice += doBestTotalPrice(length - newLength);
      }
      if (newTotalPrice > bestYetTotalPrice) {
        bestYetTotalPrice = newTotalPrice;
      }
    }
    Log.leave(bestYetTotalPrice);
    return bestYetTotalPrice;
  }

  /**
   * new SimpleTotalPriceOptimizer(priceList) returns a new simple total price
   * optimizer.
   * It uses priceList to determine the price for sub-bars.
   * @param priceList the price list for sub-bars
   */
  public SimpleTotalPriceOptimizer(PriceList priceList) {
    this.priceList = priceList;
  }

  /**
   * tpo.bestTotalPrice(length) returns the best total price that can be charged
   * for sub-bars cut from a bar of length length.
   * @param length the bar length
   * @return the best total price
   */
  @Override
  public double bestTotalPrice(int length) {
    Log.reset();
    return doBestTotalPrice(length);
  }

  /**
   * tpo.toString() returns a string description for tpo.
   * @return a string description
   */
  @Override
  public String toString() {
    return "Simple Total Price Optimizer";
  }

  /**
   * @param args ignored
   */
  public static void main(String[] args) {
    Log.setTraceCalls(false);
    PriceList priceList = new PriceList(1.0, 10);
    Log.log(priceList.toString());
    Log.log();
    SimpleTotalPriceOptimizer optimizer =
            new SimpleTotalPriceOptimizer(priceList);
    Log.log(optimizer);
    Log.log("Best Total Prices:");
    // To keep total run time under 30 seconds, use maxLength <= 28.
    // Stick to maxLength <= 20 if tracing calls.
    int maxLength = 20;
    for (int length = 1; length <= maxLength; length++) {
      Log.log(length + ": " + optimizer.bestTotalPrice(length)
              + ", call count = " + Log.getCallCount());
    }
    System.out.println(Log.getContents());
  }

}
