package totalprice;

import java.util.ArrayList;
import java.util.List;

/**
 * A Solution is a solution of the problem of optimizing the total price of
 * sub-bars of a bar.
 * It is created with parameters to specify the price list to be used and the
 * bar length.
 * When created, the entire bar is the only sub-bar.
 * This is not usually an optimal solution.
 * It can be refined by invoking refineSolution() with parameters specifying
 * a first cut length and an optimized solution for the remainder of the bar.
 */
public class Solution {

  private PriceList priceList;
  private int length;
  private double totalPrice;
  private int firstLength;
  private Solution subproblemSolution;
  private String subbarLengthsString;

  /**
   * new Solution(priceList, length) returns a new starting total price
   * optimization solution for a bar of length length, using priceList to
   * determine prices for sub-bars.
   * The initial solution does not cut the bar.
   * The solution is refined with later calls to reviseSolution().
   * @param length
   */
  public Solution(PriceList priceList, int length) {
    this.priceList = priceList;
    this.length = length;
    totalPrice = this.priceList.getPrice(length);
    firstLength = length;
    subproblemSolution = null;
  }

  /**
   * sol.refineSolution(firstLength, subproblemSolution) revises sol if the
   * solution consisting of a bar of length firstLength and the subbars of
   * subProblemSolution yields a better total price.
   * @param firstLength the first sub-bar length
   * @param subproblemSolution a solution for the remainder of the bar
   */
  public void reviseSolution(int firstLength, Solution subproblemSolution) {
    if (firstLength + subproblemSolution.getLength() != length) {
      throw new IllegalArgumentException("Lengths do not add up");
    }
    double newTotalPrice = priceList.getPrice(firstLength);
    if (subproblemSolution != null) {
      newTotalPrice += subproblemSolution.getTotalPrice();
    }
    if (newTotalPrice > totalPrice) {
      totalPrice = newTotalPrice;
      this.firstLength = firstLength;
      this.subproblemSolution = subproblemSolution;
    }
  }

  /**
   * sol.getLength() returns the length of the bar that sol solves.
   * @return the bar length
   */
  public int getLength() {
    return length;
  }

  /**
   * sol.getTotalPrice() returns the total price for the sub-bars of sol.
   * @return the total price
   */
  public double getTotalPrice() {
    return totalPrice;
  }

  /**
   * sol.getSubbarLengthsString() returns a comma-separated list of the subbar
   * lengths of sol.
   * @return a comma-separated length list
   */
  public String getSubbarLengthsString() {
    if (subbarLengthsString == null) {
      subbarLengthsString = Integer.toString(firstLength);
      if (subproblemSolution != null) {
        subbarLengthsString +=
                ", " + subproblemSolution.getSubbarLengthsString();
      }
    }
    return subbarLengthsString;
  }

  /**
   * sol.addSubbarLengths(subbarLengthList) adds the sub-bar lengths of sol to
   * subbarLengthList.
   * @param subbarLengthList a list
   */
  public void addSubbarLengths(List<Integer> subbarLengthList) {
    subbarLengthList.add(firstLength);
    if (subproblemSolution != null) {
      subproblemSolution.addSubbarLengths(subbarLengthList);
    }
  }

  /**
   * sol.createSubbarLengthList() returns a sub-bar length list for sol.
   * @return
   */
  public List<Integer> createSubbarLengthList() {
    List<Integer> subbarLengthList = new ArrayList<Integer>();
    addSubbarLengths(subbarLengthList);
    return subbarLengthList;
  }

  /**
   * sol.toString returns a description of sol.
   * @return a description
   */
  @Override
  public String toString() {
    return "For a bar of length " + length + ", total price = " + totalPrice
            + ".  Sub-bar lengths are " + getSubbarLengthsString();
  }

}
